﻿--####################################################################################
--####################################################################################
--Main Sub
--####################################################################################
--####################################################################################
--Dependencies: none

--Bindings.xml
BINDING_HEADER_COMPRESSMACRO		= "CompressMacro";
BINDING_NAME_COMPRESSMACRO_EDIT		= "Show main window";

local CM_frame						= nil; --Self reference(s) used for RegisterEvent/UnregisterEvent
--CM_GameVersion					= nil; --UIversion number of the game for quick lookup for other parts of the code for compatibility
	--if (CM_GameVersion < 50300) then
		--Before patch 5.3.0

local CM_addon_version = GetAddOnMetadata("CompressMacro", "Version");	--Version number for the addon

--Local pointers to global functions
local tonumber	= tonumber;
local tostring	= tostring;
local pairs		= pairs;
local strlen	= strlen;
local strsub	= strsub;
local strfind	= strfind;
local strlower	= strlower;
local strupper	= strupper;
local gsub		= gsub;
local strtrim	= strtrim;	--string.trim
local sort		= sort;

local print				= print;
local IsMouseButtonDown	= IsMouseButtonDown;
local GetMouseFocus		= GetMouseFocus;


--####################################################################################
--####################################################################################
--Event Handling and Macro Scriptmethod
--####################################################################################

CompressMacro			= {};	--Global declaration
CompressMacro.__index	= CompressMacro;


--Local variables that cache stuff so we dont have to recreate large objects
local cache_addonLoaded			= false;	--flag to tell us wehter the addon is loaded (so we dont reinit when instancing etc)


--OnLoad Event
function CompressMacro:OnLoad(s)
	--We here only register for the events that we want to listen to, CompressMacro:OnEvent() handles the various events
	CM_frame = s; --save it for later reference
	s:RegisterEvent("PLAYER_ENTERING_WORLD");
	return nil;
end


--Handles the events for the addon
function CompressMacro:OnEvent(s, event, ...)
	if (event == "PLAYER_ENTERING_WORLD") then
		--Startup
		--local ___version, ___internalVersion, ___thedate, ___uiVersion = GetBuildInfo();
		--CM_GameVersion = ___uiVersion; --UI version number used for compatibility checks between game versions.

		if (cache_addonLoaded) then return nil end --Stop loading if this has already been done

		--Add the slash command
		SLASH_CompressMacro1 = "/compressmacro";
		--SLASH_CompressMacro2 = "/compress";
		SlashCmdList["CompressMacro"] = function(cmd) return self:Slash(cmd) end;

		--Show welcome message
		self:msg("CompressMacro loaded ("..CM_addon_version.."). Use "..SLASH_CompressMacro1.." to display window.");--" or "..SLASH_CompressMacro2..

		--Register with LibDataBroker
		if (LibStub ~= nil) then
			local lib = LibStub:GetLibrary("LibDataBroker-1.1"); --https://github.com/tekkub/libdatabroker-1-1/
			lib:NewDataObject("CompressMacro", {type="launcher", icon="254884", text="CompressMacro", OnClick=function(clickedframe, button) CompressMacro:Slash(""); end,} );--OLD: Interface\\Icons\\INV_Sigil_Mimiron NEW: 254884
		end --if LibStub

		s:UnregisterEvent("PLAYER_ENTERING_WORLD"); --make sure this event isnt triggered again
		cache_addonLoaded = true; --set flag so that this stuff is only done once
	end

	return nil;
end


--####################################################################################
--####################################################################################
--Slash-handler
--####################################################################################


--Handler for slash commands
function CompressMacro:Slash(cmd)
	if (cmd == nil) then cmd = ""; end
	cmd = strtrim(strlower(cmd)); --tweak the input string

	local frmMain = CompressMacroXML_Edit_Frame;
	local frmList = CompressMacroXML_Help_Frame;

	if (cmd == "list") then
		--Show the list window
		if (frmList:IsVisible() == 1) then
			frmList:Hide();
		else
			if (frmMain:IsVisible() ~= 1) then frmMain:Show(); end --Dont show list window alone
			frmList:Show();
			frmList:Raise();
		end --if IsVisible
	else
		--Show the edit window
		if (frmMain:IsVisible() == 1) then
			frmMain:Hide();
			if (frmList:IsVisible() == 1) then frmList:Hide(); end --Hide list window too
		else
			local edt = CompressMacroXML_Edit_EditBox1;
			--edt:SetText(txt);			--Refresh text from memory
			edt:SetCursorPosition(0);	--Set cursor position to the start of the text
			--edt:SetCursorPosition(edt:GetNumLetters()); --Set cursor position to the end of the text
			--scr:SetVerticalScroll(0) --Reset the scrollbar to the top

			frmMain:Show();
			frmMain:Raise();
		end --if IsVisible
	end--if cmd

	return nil;
end


--####################################################################################
--####################################################################################
--Window Frame handling
--####################################################################################


--Refresh the textcounter
function CompressMacro:Edit_TextChanged(objEdit, isUserInput, outputLabel)
	--print("CompressMacroXML_Edit_TextChanged");

	--objEdit is a EditBox
	local m = tostring(objEdit:GetMaxLetters());
	local c = tostring(objEdit:GetNumLetters());

	if (tonumber(c) >= tonumber(m)) then
		message("You have reached the maximum number of characters.");
	end

	--outputLabel is a FontString
	--outputLabel:SetText(c.."/"..m.." characters.");
	outputLabel:SetText(c.." characters.");
	return nil;
end


--Toggle a frame between beign frontmost or behind any other frames that the user clicks on
function CompressMacro:ToggleFrameStrata(s, forceFront)
	--if true then return nil end

	if forceFront then
		--Frame has recived manual focus (mouseclick) and needs to be put topmost
		s:SetFrameStrata("HIGH");
		s:SetToplevel(true);
		--print("moved front");
	else
		--Regular polling called from OnUpdate() to whether we shall change focus or not
		if (IsMouseButtonDown() == nil) then return nil end		--if the mouse button is not pressed down, then skip the rest
		local objMouseFrame		= GetMouseFocus();				--get the name of the frame that the mouse is currently focusing on...
		if (objMouseFrame == nil or objMouseFrame["GetFrameStrata"] == nil) then return nil end			--GetMouseFocus will some cases return nil
		local strMouseStrata	= objMouseFrame:GetFrameStrata() or "BACKGROUND";	--keep the frame's strata until later...
		local strMainFrame		= s:GetName();

		repeat				--We do a loop where each element and their parent(s) is checked and we see if that eventually matches our frame,
			if (objMouseFrame ~= nil) then
				--print("  loop: "..objMouseFrame:GetName());
				if (objMouseFrame:GetName() == strMainFrame) then return nil end --if we find our mainform then we skip the rest
			end--if
			objMouseFrame = objMouseFrame:GetParent();
		until (objMouseFrame == nil)

		--If we get down here then then:
		--			1. A mouse button is pressed down
		--			2. None of the controls on the frame is selected
		--We set the frame's strata to the lowest we can...

		--We set the frame's strata to that level just below what the mouse is focused on...
		if 	   strMouseStrata=="TOOLTIP"			then strMouseStrata="FULLSCREEN_DIALOG"
		elseif strMouseStrata=="FULLSCREEN_DIALOG"	then strMouseStrata="FULLSCREEN"
		elseif strMouseStrata=="FULLSCREEN"			then strMouseStrata="DIALOG"
		elseif strMouseStrata=="DIALOG"				then strMouseStrata="HIGH"
		elseif strMouseStrata=="HIGH"				then strMouseStrata="MEDIUM"
		elseif strMouseStrata=="MEDIUM"				then strMouseStrata="LOW"
		elseif strMouseStrata=="LOW"				then strMouseStrata="BACKGROUND"
		else										     strMouseStrata="BACKGROUND"
		end--if

		s:SetFrameStrata(strMouseStrata);
		s:SetToplevel(false);
		--print("moved back ");
	end

	return nil;
end


function CompressMacro:SettingsFrame_OnLoad(panel, title)
	title:SetText("CompressMacro "..tostring(CM_addon_version).." ");--<FontString> containing the title & version number

	-- Set the name of the Panel
	panel.name = "CompressMacro";

	--We dont use any of these
	--panel.okay	= function (self) CompressMacro:SettingsFrame_Ok(); end;
	--panel.cancel	= function (self) CompressMacro:SettingsFrame_Cancel(); end;
	panel.default	= function (self) print("There are no settings for CompressMacro"); end;
	--panel.refresh	= function (self) CompressMacro:SettingsFrame_Refresh(); end;

	-- Add the panel to the Interface Options
	InterfaceOptions_AddCategory(panel);
end


--####################################################################################
--####################################################################################
--Development, Testing, Support
--####################################################################################


--Outputs a message into the chat frame
function CompressMacro:msg(n)
	print("|cFF0A8AFFCompressMacro|cFFFFFFFF:|r " .. n);
	return true;
end


--Returns true/false whether a string starts with the given item
function CompressMacro:startsWith(str, item)
	if (str == nil or str == "" or item == nil or item == "" or strlen(item) > strlen(str)) then return false end

	local sPos1 = strfind(str, item, 1, true); --plain find starting at pos 1 in the string
	if (sPos1 == 1) then return true end --if the string starts at the first position in the string then we accept it
	return false;
end

--Replaces any occurences of 'old' with 'new', case-insensitive
function CompressMacro:replaceNoCase(str, old, new)
	local _strlower	= strlower; --local fpointer

	old = _strlower(old);
	if (old == _strlower(new)) then return str; end--if we didnt have this then we would get into an infinite loop

	local _strfind	= strfind; --local fpointer
	local _strsub	= strsub;
	local _gsub		= gsub;

	while true do
		local start, finish = _strfind(_strlower(str), old, 1, true); --plain string search with lowercase version to get start and finish location
		if (start == nil) then break; end
		local value	= _strsub(str, start, finish);			--pull out exact string as it is written
		value		= self:escapeMagicalCharacters(value);	--escape any magical characters so that they are seen as literal strings
		str = _gsub(str, value, new);						--replace exact string with new value
	end--while

	return str;
end

local CONST_escapeMagicalCharacters = {"(",")",".","%","+","-","*","?","[","]","^","$"}; --Hardcoded, these are the magical characters that have special meaning when it comes to LUA patterns, by adding % infront of them we escape them

--Replaces any occurences of ( ) . % + - * ? [ ] ^ $ by adding a % ahead of it
function CompressMacro:escapeMagicalCharacters(str)
	local res = "";
	local _strsub = strsub; --local fpointer
	for i = 1, strlen(str) do
		local char = _strsub(str,i,i);
		for j = 1, #CONST_escapeMagicalCharacters do
			local esc = CONST_escapeMagicalCharacters[j];
			if (char == esc) then
				char = "%"..char;
				break;
			end--if
		end--for j
		res = res..char;
	end--for i

	return res;
end


--####################################################################################
--####################################################################################
--Macro Compression Code
--####################################################################################


local cache_SlashAll		= nil; --Table with all slash commands ingame {key="/whisper", value="CHAT_WHISPER"}
local cache_SlashSmallest	= nil; --Table with the shortest slash command in each family {key="CHAT_WHISPER", value="/w"}
local cache_DisplaySmallest	= nil; --String with the formatted text to display
local cache_DisplayFamily	= nil;

local CONST_GCTimespan	= 30;
local TimeSinceLastGC	= time(); --timestamp when :ParseText() last did a GC


--Parse text and update text labels
function CompressMacro:ParseText(Edit1, Edit2)
	local rawText = Edit1:GetText();
	local newText = self:Compress_Macro(rawText);
	Edit2:SetText(newText);

	local d = difftime(time(), TimeSinceLastGC); --Garbage Collect every n seconds to prevent bleeding
	if (d > CONST_GCTimespan) then collectgarbage("collect"); TimeSinceLastGC = time(); end
	return nil;
end


--Populate editbox with list of slashcommands
function CompressMacro:getSlashText(Edit1, listType)
	--Populate the edit box with the text
	local output = "";

	if (listType == "small") then
		if (cache_DisplaySmallest ~= nil) then
			output = cache_DisplaySmallest; --fetch from cache
		else
			local tbl, count = self:getSlashList(listType);
			output = "Showing the shortest slashcommand for each family-grouping.\n\nTotal number of commands: "..tostring(count).."\n";
			for i=1, #tbl do
				output = output.."\n"..tbl[i];
			end--for i
			cache_DisplaySmallest = output; --store in cache for later lookups
			if (cache_DisplaySmallest ~= nil and cache_DisplayFamily ~= nil) then self["getSlashList"] = nil; end --Cleanup
			collectgarbage("collect"); --force a garbage collection
		end--if

	elseif (listType == "family") then
		if (cache_DisplayFamily ~= nil) then
			output = cache_DisplayFamily; --fetch from cache
		else
			local tbl, count = self:getSlashList(listType);
			output = "Showing all slashcommands, grouped by family.\n\nTotal number of commands: "..tostring(count).."\n";
			for i=1, #tbl do
				output = output.."\n"..tbl[i];
			end--for i
			cache_DisplayFamily = output; --store in cache for later lookups
			if (cache_DisplaySmallest ~= nil and cache_DisplayFamily ~= nil) then self["getSlashList"] = nil; end --Cleanup
			collectgarbage("collect"); --force a garbage collection
		end--if
	end--if listType


	Edit1:SetText(output);
	Edit1:SetCursorPosition(0);	--Set cursor position to the start of the text
	return nil;
end


--Returns and array with sorted slashcommands and a total count of commands
function CompressMacro:getSlashList(listType)
	if (cache_SlashAll == nil) then
		cache_SlashAll, cache_SlashSmallest = self:declareSlashList(); --initialize variables and cache them in this file.
		self["declareSlashList"] = nil; --Cleanup
		--collectgarbage("collect"); --force a garbage collection
	end

	--listType = "small"	: only the smallest slashcommands, alphabetically sorted
	--listType = "family"	: grouped by family of slashcommands, alphabetically sorted inside the family and for the whole list.

	local tbl, count = {}, 0;
	local _pairs	= pairs; --local fpointer
	local _sort		= sort;

	if (listType == "small") then
		local tblSmall = cache_SlashSmallest;
		for family,slash in _pairs(tblSmall) do
			tbl[#tbl+1] = slash;--.."     "..family
		end--for
		_sort(tbl); --Alphabetical sort
		count = #tbl;


	elseif listType == "family" then
		--get tblsmall and sort alphabetical after that...
		local tblAll	= cache_SlashAll;
		local tblSmall	= cache_SlashSmallest;
		local tblTmp	= {};
		local strSpacer	= "";

		--Create a sorted table of the shortest commands
		for family,slash in _pairs(tblSmall) do
			tblTmp[#tblTmp+1] = slash;--.."     "..family
		end--for
		_sort(tblTmp); --Alphabetical sort

		count = 0;
		--Iterate though cache_SlashAll and get the rest of the current family
		for i=1,#tblTmp do
			local currSlash  = tblTmp[i];
			local currFamily = "";
			--Find the family name for the current slash
			for family,slash in _pairs(tblSmall) do
				if (slash == currSlash) then
					currFamily = family;
					break;
				end--if
			end--for tblSmall

			local currTbl = {};
			currTbl[#currTbl+1] = currSlash;

			--find each family member
			for slash,family in _pairs(tblAll) do
				if (currFamily == family) then
					currTbl[#currTbl+1] = slash;
				end
			end--for tblAll
			_sort(currTbl); --Sort the family group

			--Add the sorted family to the rest of the result
			for j=1, #currTbl do
				tbl[#tbl+1] = currTbl[j];
				count = count +1; --increment this way so we dont count spacers
			end--for j
			tbl[#tbl+1] = strSpacer; --add a spacer in between groupings
		end--for i
	end--if listType

	return tbl, count;
end


--Compress all slashcommands in a text to their minimum size (i.e '/whisper' to '/w' etc)
function CompressMacro:Compress_Macro(macroBody)
	if (macroBody == nil or macroBody == "") then return macroBody end;

	--We don't declare and waste Memory & CPU on this until someone actually requires it. We prefer using only a little bit of memory from startup to hold the function rather than having to traverse the entire _G in search for slash commands at startup.
	if (cache_SlashAll == nil) then
		--local startTime = debugprofilestop();
		cache_SlashAll, cache_SlashSmallest = self:declareSlashList(); --initialize variables and cache them in this file.
		self["declareSlashList"] = nil; --Cleanup
		--print(format("CompressMacro_Macro:declareSlashList() %f ms", debugprofilestop()-startTime));
		collectgarbage("collect"); --force a garbage collection
	end

	--Traverse the macroBody and replace anything we find with the smallest string
	local _pairs = pairs; --local fpointer

	local tblAll	= cache_SlashAll;
	local tblSmall	= cache_SlashSmallest;

	--All strings are lowercased already when declared.
	--We also added a space at the end of the slashcommands so that we dont have to do here ('/whisper ')
	for slash,family in _pairs(tblAll) do
		local smallest = tblSmall[family];
		if (slash ~= smallest) then
			macroBody = self:replaceNoCase(macroBody, slash, smallest);
		end--if
	end--for k,v

	return strtrim(macroBody);
end


--Traverses _G in search for _G["SLASH_*"] keys. Then it groups each command into a family depending on the argument name
function CompressMacro:declareSlashList()

	local strlower	= strlower; --local fpointer
	local strtrim	= strtrim;
	local strupper	= strupper;
	local pairs		= pairs;
	local strsub	= strsub;
	local strlen	= strlen;
	local type		= type;

	local _G		= _G;
	local strKey	= "SLASH_";
	local strValue	= "/";
	local tblAll	= {}; --Table with all slash commands ingame {key="/whisper", value="CHAT_WHISPER"}
	local tblSmall	= {}; --Table with the shortest slash command in each family {key="CHAT_WHISPER", value="/w"}

	for k,v in pairs(_G) do --go though _G and find all the "SLASH_" entries
		if (type(k) == "string" and self:startsWith(k, strKey) == true and type(v) == "string" and self:startsWith(v, strValue)) then --key must start with SLASH_ and the value must start with /
			local slash	= strtrim(strlower(v)).." ";			--/guild			(/g /guild are different but they will have the same family name). We add a space at the end so that we dont have to do it later in CompressMacro()
			local family= strtrim(strupper(k));					--GUILD_COMMAND7	(familyname: all related slash commands got this incommon)
			family		= strsub(family, 1, strlen(family)-1);	--strip away the digit at the end of the key

			--Check if this exact slash already exists in the table
			if (tblAll[slash] == nil) then tblAll[slash] = family; end --Add to table if unique

			--Replace with the shortest slash out of those in the same family
			local tmp = tblSmall[family];
			if (tmp == nil)									then tblSmall[family] = slash; end
			if (tmp ~= nil and strlen(tmp) > strlen(slash))	then tblSmall[family] = slash; end
		end --if startsWith
	end--for k,v

	--We strip out the shortest slashes from the All table, that way we can save alot of space since these tables tend to be big.
	--i.e '/whisper ' we want to compress to '/w '. But '/w ' can not be compressed any further so we dont need to keep track of it.
	local tmp = {};
	for slash,family in pairs(tblAll) do
		if (tblSmall[family] ~= slash) then tmp[slash] = family; end
	end--for tblAll
	tblAll = tmp;

	return tblAll, tblSmall;
end


--local startTime = debugprofilestop();
--print(format("CompressMacro %f ms", debugprofilestop()-startTime));
--####################################################################################
--####################################################################################

--[[
--Will create a new default set of debugging macro's
function beta_doMacro()
	local body  = {}; --macro body and name
	body["debug1"] = "/reload";
	body["debug2"] = "/dump foo()";
	body["debug3"] = "/dump bar()";

	local texture = {}; --macro texture
	texture["debug1"] = "INV_Sigil_Thorim";
	texture["debug2"] = "INV_Sigil_Mimiron";
	texture["debug3"] = "ACHIEVEMENT_GUILDPERK_FASTTRACK_RANK2";

	local toolbarslot = {}; --actionbar index 1-12 for first toolbar, 13-X for second and so on
	toolbarslot["debug1"] = 14;
	toolbarslot["debug2"] = 15;
	toolbarslot["debug3"] = 16;

	for k,v in pairs(body) do
		if (GetMacroIndexByName(k) == 0) then
			CreateMacro(k, texture[k], v, nil); --Always created as a shared macro (not per character)
			print("macro created "..tostring(k));

			if (GetActionInfo(toolbarslot[k]) == nil) then --if toolbar is not empty...
				PickupMacro(k); --place macro on cursor
				PlaceAction(toolbarslot[k]); --place cursor item on toolbar
			end--if GetActionInfo
		end--if GetMacroIndexByName
	end--k,v
end--]]--
